// -*- js-indent-level: 2; js-switch-indent-offset: 2 -*-
"use strict";

function log(m) {
  jsShim.logMessage(''+m);
}

/**
 * Wrap the NPAPI plugin's callWithObjectAddress method and
 * propagate exceptions across the call boundary.
 */
function callWithObjectAddress(callback, oracle, obj) {
  var caughtException = undefined;
  var ret = oracle.callWithObjectAddress(
    function(addr) {
      try {
        callback(addr);
      } catch(ex) {
        caughtException = ex;
      }
    }, obj);

  if(caughtException !== undefined) {
    throw caughtException;
  }

  if(jsShim.isError(ret)) {
    throw ret;
  }
}

function callWithNpFunctionsAddress(callback, oracle) {
  var caughtException = undefined;
  var ret = oracle.callWithNpFunctionsAddress(
    function(addr) {
      try {
        callback(addr);
      } catch(ex) {
        caughtException = ex;
      }
    });

  if(caughtException !== undefined) {
    throw caughtException;
  }

  if(jsShim.isError(ret)) {
    throw ret;
  }
}

function enumPlugins() {
  log('Enumerating navigator.plugins');
  var plugins = navigator.plugins;
  var n = plugins.length;
  for(var i = 0; i < n; ++i) {
    log('plugin: ' + plugins[i].name);
  }
}

function doInstallPointerPlugin() {
  enumPlugins();
  var ret = jsShim.installPointerPlugin();
  if(jsShim.isError(ret)) {
    throw ret;
  }
}

function reportError(ex) {
  if(jsShim.isError(ex)) {
    jsShim.onError(ex);
  } else {
    jsShim.onErrorString(ex+'');
  }
}

var timesPolled = 0;
var pollInterval = 1;

function pollForPluginReady() {
  timesPolled += 1;
  var oracle = document.embeds[0];
  if(oracle.callWithObjectAddress) {
    handleHackPhase_1('hackinstall');
  } else if(timesPolled > 1000) {
    reportError('timed out waiting for plugin to load');
  } else {
    window.setTimeout(pollForPluginReady, pollInterval);
    if(pollInterval < 256) {
      pollInterval *= 2;
    }
  }
}

function loadPlugin() {
  var plugin = document.createElement('embed');
  plugin.type = 'application/x-facebook-hack';
  plugin.hidden = true;
  document.body.appendChild(plugin);

  // We need to wait for the browser to get around to loading the
  // embed content (our plugin) so that we can call into our plugin
  // with JS.  You might think we could just listen for a load event,
  // but at least on GB, that event seems to never fire.
  //
  // Instead, let's poll. :-(
  //

  pollForPluginReady();
}

function doFix() {
  var oracle = document.embeds[0];

  callWithNpFunctionsAddress(
    function(addr) {
      var ret = jsShim.fixObject('npn', addr);
      if(jsShim.isError(ret)) {
        throw ret;
      }
    }, oracle);

  function fix(what, obj) {
    callWithObjectAddress(
      function(addr) {
        var ret = jsShim.fixObject(what, addr);
        if(jsShim.isError(ret)) {
          throw ret;
        }
      }, oracle, obj);
  }

  fix('window', window);
  fix('window.open', window.open);
  fix('window.location', window.location);
  fix('location.assign', window.location.assign);
  fix('location.replace', window.location.replace);
  gotoPhase('hacktest');
}

function doLoadTest(cb) {
  var iframe = document.createElement('iframe');
  iframe.src = 'about:blank';
  iframe.name = 'inner';
  iframe.onload = function() {
    cb();
    return true;
  };
  document.body.appendChild(iframe);
}

var BLOCKED_REGEX =
  /^Error: (unsafe DOM call blocked|unsafe mutation of .+ detected)$/;

function expectSecurityBlock(fn) {
  var src = fn.toString().replace(/[ \t\n\r]+/g, ' ');
  var caught = false;
  log('testing that security fix blocks this call: ' + src);
  try {
    fn();
  } catch(ex) {
    if((ex+'').match(BLOCKED_REGEX)) {
      caught = true;
    } else {
      throw ex;
    }
  }

  if(!caught) {
    throw new Error('call to' + src + ': ' +
                    'expected security block, ' +
                    'but call unexpectedly succeeded');
  }

  log('... success');
}

function expectSuccess(fn) {
  var src = fn.toString().replace(/[ \t\n\r]+/g, ' ');
  log('expecting this call to succeed: ' + src);
  fn();
}

function doTest() {
  var iframe = event.target;
  iframe.onload = null;
  var w = iframe.contentWindow;
  var exploitString = '\u0000javascript:2+2';

  expectSecurityBlock(function() {w.location.replace(exploitString);});
  expectSecurityBlock(function() {w.location.assign(exploitString);});
  expectSecurityBlock(function() {window.open(exploitString, null, iframe.name);});
  expectSecurityBlock(function() {w.location.href = exploitString;});
  expectSecurityBlock(function() {w.location.href = new String(exploitString);});
  expectSecurityBlock(function() {w.location['href'] = new String(exploitString);});
  expectSecurityBlock(function() {w.location.protocol = exploitString;});
  expectSecurityBlock(function() {w.location.host = exploitString;});
  expectSecurityBlock(function() {w.location.port = exploitString;});
  expectSecurityBlock(function() {w.location.pathname = exploitString;});
  expectSecurityBlock(function() {w.location.search = exploitString;});
  expectSecurityBlock(function() {w.location.hash = exploitString;});

  expectSuccess(function() {w.location.replace('about:blank');});
  expectSuccess(function() {w.location.assign('about:blank');});
  expectSuccess(function() {window.open('about:blank', null, iframe.name);});
  expectSuccess(function() {w.location.href = 'about:blank';});
  expectSuccess(function() {w.location.href = new String('about:blank');});
  expectSuccess(function() {w.location['href'] = new String('about:blank');});
  expectSuccess(function() {w.location.protocol = 'about';});
  expectSuccess(function() {w.location.host = 'localhost';});
  expectSuccess(function() {w.location.port = '80';});
  expectSuccess(function() {w.location.pathname = '/test';});
  expectSuccess(function() {w.location.search = '111';});
  expectSuccess(function() {w.location.hash = '222';});

  w.stop();
}

function gotoPhase(phase) {
  log('switching to phase [' + phase + ']');
  var baseUrl = location.href.replace(/[?#].*$/, '');
  var newUrl = baseUrl + '?' + phase;
  location.replace(newUrl);
}

function handleHackPhase_1(phase) {
  try {
    log('current phase: [' + phase + ']');
    switch(phase) {
      case "":
        doInstallPointerPlugin();
        gotoPhase('loadPlugin');
        break;
      case 'loadPlugin':
        enumPlugins();
        loadPlugin();
        break;
      case 'hackinstall':
        doFix();
        break;
      case 'hacktest':
        doLoadTest(function() {
          handleHackPhase_1('hacktestLoaded');
        });
        break;
      case 'hacktestLoaded':
        doTest();
        gotoPhase('success');
        break;
      case 'success':
        jsShim.onSuccess();
        break;
      default:
        throw new Error('unknown phase: [' + phase + ']');
    }
  } catch(ex) {
    reportError(ex);
  }
}

function handleHackPhase() {
  handleHackPhase_1(location.search.replace('?', ''));
}

function onLoadHandler() {
  log('onload of [' + window.location + ']');
  handleHackPhase();
  return true;
}
